@@ -1,24 +1,38 @@
 from contextlib import contextmanager
-import collections
+import attr
 import socket
 import yaml
+import re
+
+import six
 
 
-class Job(collections.namedtuple('Job', ['job_id', 'job_data'])):
-    """Structure holding a job returned from Beanstalk"""
-    pass
+@attr.s(frozen=True)
+class Job(object):
+    """Structure holding a job returned from Beanstalk
+
+    :ivar job_id: Opaque identifier for the job (to be passed to :func:`BeanstalkClient.release_job()` or
+        :func:`BeanstalkClient.stats_job()`).
+
+    :ivar job_data: Blob of the data. str if :attr:`BeanstalkClient.auto_decode` is True; otherwise bytes
+    """
+
+    job_id = attr.ib(validator=attr.validators.instance_of(int))
+
+    job_data = attr.ib()
 
 
+if getattr(attr, "__version_info__", (0,)) >= (19, 2):
+    _attrs_kwargs = dict(eq=True)
+else:
+    _attrs_kwargs = dict(cmp=True)
+
+
+@attr.s(frozen=True, hash=True, **_attrs_kwargs)
 class BeanstalkError(Exception):
-    """Common error raised when something goes wron with beanstalk"""
-    def __init__(self, message):
-        self.message = message.decode('ascii')
+    """Common error raised when something goes wrong with beanstalk"""
 
-    def __repr__(self):
-        return '{0}({1!r})'.format(self.__class__.__name__, self.message)  # pragma: no cover
-
-    def __str__(self):
-        return repr(self)  # pragma: no cover
+    message = attr.ib(converter=lambda m: m.decode('ascii'))
 
 
 def yaml_load(fo):
@@ -29,6 +43,31 @@ def yaml_load(fo):
         return yaml.safe_load(fo)
 
 
+@attr.s(frozen=True)
+class BeanstalkInsertingProxy(object):
+    """Proxy object yielded by :func:`BeanstalkClient.using()`"""
+
+    beanstalk_client = attr.ib()
+    tube = attr.ib()
+
+    def put_job(self, data, pri=65536, delay=0, ttr=120):
+        """Method to insert a job into the tube selected with :func:`BeanstalkClient.using`.
+
+        :param data: Job body
+        :type data: Text (either str which will be encoded as utf-8, or bytes which are already utf-8
+        :param pri: Priority for the job
+        :type pri: int
+        :param delay: Delay in seconds before the job should be placed on the ready queue
+        :type delay: int
+        :param ttr: Time to reserve (how long a worker may work on this job before we assume the worker is blocked
+            and give the job to another worker
+        :type ttr: int
+        """
+
+        self.beanstalk_client.use(self.tube)
+        return self.beanstalk_client.put_job(data=data, pri=pri, delay=delay, ttr=ttr)
+
+
 class BeanstalkClient(object):
     """Simple wrapper around the Beanstalk API.
 
@@ -54,34 +93,77 @@ class BeanstalkClient(object):
         self.host = host
         self.port = port
         self.socket_timeout = socket_timeout
-        self.socket = None
-        self.watchlist = ['default']
+        self._reset_state()
+        self.desired_tube = 'default'
+        self.desired_watchlist = set(['default'])
+        self.auto_decode = auto_decode
+
+    @classmethod
+    def from_uri(cls, uri, socket_timeout=None, auto_decode=False):
+        """Construct a synchronous Beanstalk Client from a URI.
+
+        The URI may be of the form beanstalk://host:port or beanstalkd://host:port
+
+        IPv6 literals must be wrapped in brackets as per RFC 2732.
+        """
+        parts = six.moves.urllib.parse.urlparse(uri)
+        if parts.scheme.lower() not in ('beanstalk', 'beanstalkd'):
+            raise ValueError('Invalid scheme %s' % parts.scheme)
+        ipv6_md = re.match(r'^\[([0-9a-fA-F:]+)\](:[0-9]+)?$', parts.netloc)
+        if ipv6_md:
+            host = ipv6_md.group(1)
+            port = ipv6_md.group(2) or '11300'
+            port = port.lstrip(':')
+        elif ':' in parts.netloc:
+            host, port = parts.netloc.rsplit(':', 1)
+        else:
+            host = parts.netloc
+            port = 11300
+        port = int(port)
+        return cls(host, port, socket_timeout=socket_timeout, auto_decode=auto_decode)
+
+    def _reset_state(self):
+        self._watchlist = set(['default'])
         self.current_tube = 'default'
         self.initial_watch = True
-        self.auto_decode = auto_decode
+        self.socket = None
 
     def __repr__(self):
         return '{0}({1!r}, {2!r})'.format(self.__class__.__name__, self.host, self.port)  # pragma: no cover
 
     def __str__(self):
         return '{0} - watching:{1}, current:{2}'.format(  # pragma: no cover
-            repr(self), self.watchlist, self.current_tube  # pragma: no cover
+            repr(self), self._watchlist, self.current_tube  # pragma: no cover
         )  # pragma: no cover
 
     @property
     def _socket(self):
         if self.socket is None:
             self.socket = socket.create_connection((self.host, self.port), timeout=self.socket_timeout)
+            self._re_establish_use_watch()
         return self.socket
 
+    def _re_establish_use_watch(self):
+        """Call after a close/re-connect.
+
+        Automatically re-establishes the USE and WATCH configs previously setup.
+        """
+        if self.current_tube != self.desired_tube:
+            self.use(self.desired_tube)
+        if self._watchlist != self.desired_watchlist:
+            self.watchlist = self.desired_watchlist
+
     def close(self):
         """Close any open connection to the Beanstalk server.
 
         This object is still safe to use after calling :func:`close()` ; it will automatically reconnect
+        and re-establish any open watches / uses.
+
+        It is a logic error to close the connection while you have a reserved job
         """
         if self.socket is not None:
             self.socket.close()
-            self.socket = None
+            self._reset_state()
 
     @contextmanager
     def _sock_ctx(self):
@@ -190,7 +272,7 @@ class BeanstalkClient(object):
             return stats
 
     def put_job(self, data, pri=65536, delay=0, ttr=120):
-        """Insert a new job into the queue.
+        """Insert a new job into whatever queue is currently USEd
 
         :param data: Job body
         :type data: Text (either str which will be encoded as utf-8, or bytes which are already utf-8
@@ -201,6 +283,14 @@ class BeanstalkClient(object):
         :param ttr: Time to reserve (how long a worker may work on this job before we assume the worker is blocked
             and give the job to another worker
         :type ttr: int
+
+        .. seealso::
+
+           :func:`put_job_into()`
+              Put a job into a specific tube
+
+           :func:`using()`
+              Insert a job using an external guard
         """
         with self._sock_ctx() as socket:
             message = 'put {pri} {delay} {ttr} {datalen}\r\n'.format(
@@ -213,6 +303,50 @@ class BeanstalkClient(object):
             self._send_message(message, socket)
             return self._receive_id(socket)
 
+    def put_job_into(self, tube_name, data, pri=65536, delay=0, ttr=120):
+        """Insert a new job into a specific queue. Wrapper around :func:`put_job`.
+
+        :param tube_name: Tube name
+        :type tube_name: str
+        :param data: Job body
+        :type data: Text (either str which will be encoded as utf-8, or bytes which are already utf-8
+        :param pri: Priority for the job
+        :type pri: int
+        :param delay: Delay in seconds before the job should be placed on the ready queue
+        :type delay: int
+        :param ttr: Time to reserve (how long a worker may work on this job before we assume the worker is blocked
+            and give the job to another worker
+        :type ttr: int
+
+        .. seealso::
+
+           :func:`put_job()`
+              Put a job into whatever the current tube is
+
+           :func:`using()`
+              Insert a job using an external guard
+        """
+        with self.using(tube_name) as inserter:
+            return inserter.put_job(data=data, pri=pri, delay=delay, ttr=ttr)
+
+    @property
+    def watchlist(self):
+        return self._watchlist
+
+    @watchlist.setter
+    def watchlist(self, tubes):
+        """Set the watchlist to the given tubes
+
+        :param tubes: A list of tubes to watch
+
+        Automatically un-watches any tubes that are not on the target list
+        """
+        tubes = set(tubes)
+        for tube in tubes - self._watchlist:
+            self.watch(tube)
+        for tube in self._watchlist - tubes:
+            self.ignore(tube)
+
     def watch(self, tube):
         """Add the given tube to the watchlist.
 
@@ -221,16 +355,18 @@ class BeanstalkClient(object):
         Note: Initially, all connections are watching a tube named "default". If
         you manually call :func:`watch()`, we will un-watch the "default" tube.
         To keep it in your list, first call :func:`watch()` with the other tubes, then
-        call `:func:`watch()` with "default".
+        call :func:`watch()` with "default".
         """
         with self._sock_ctx() as socket:
-            self._send_message('watch {0}'.format(tube), socket)
-            self._receive_id(socket)
-            if self.initial_watch:
-                if tube != 'default':
-                    self.ignore('default')
-                self.initial_watch = False
-            self.watchlist.append(tube)
+            self.desired_watchlist.add(tube)
+            if tube not in self._watchlist:
+                self._send_message('watch {0}'.format(tube), socket)
+                self._receive_id(socket)
+                self._watchlist.add(tube)
+                if self.initial_watch:
+                    if tube != 'default':
+                        self.ignore('default')
+                    self.initial_watch = False
 
     def ignore(self, tube):
         """Remove the given tube from the watchlist.
@@ -241,11 +377,16 @@ class BeanstalkClient(object):
         to prevent the list from being empty. See :func:`watch()` for more unformation.
         """
         with self._sock_ctx() as socket:
-            if tube not in self.watchlist:
+            if tube not in self._watchlist:
                 raise KeyError(tube)
-            self._send_message('ignore {0}'.format(tube), socket)
-            self._receive_id(socket)
-            self.watchlist.remove(tube)
+            if tube != 'default':
+                self.desired_watchlist.remove(tube)
+            if tube in self._watchlist:
+                self._send_message('ignore {0}'.format(tube), socket)
+                self._receive_id(socket)
+                self._watchlist.remove(tube)
+            if not self._watchlist:
+                self._watchlist.add('default')
 
     def stats_job(self, job_id):
         """Fetch statistics about a single job
@@ -278,36 +419,37 @@ class BeanstalkClient(object):
         :type timeout: int
         """
         timeout = int(timeout)
-        if not self.watchlist:
+        if self.socket_timeout is not None:
+            if timeout >= self.socket_timeout:
+                raise ValueError('reserve_job timeout must be < socket timeout')
+        if not self._watchlist:
             raise ValueError('Select a tube or two before reserving a job')
         with self._sock_ctx() as socket:
             self._send_message('reserve-with-timeout {0}'.format(timeout), socket)
             job_id, job_data = self._receive_id_and_data_with_prefix(b'RESERVED', socket)
             return Job(job_id, job_data)
 
+    def _peek_common(self, typ):
+        """Common implementation for the peek_* functions"""
+        with self._sock_ctx() as socket:
+            self._send_message('peek-{0}'.format(typ), socket)
+            job_id, job_data = self._receive_id_and_data_with_prefix(b'FOUND', socket)
+            return Job(job_id, job_data)
+
     def peek_ready(self):
         """Peek at the job job on the ready queue.
 
         :rtype: :class:`Job`
         """
-        with self._sock_ctx() as socket:
-            self._send_message('peek-ready', socket)
-            job_id, job_data = self._receive_id_and_data_with_prefix(b'FOUND', socket)
-            return Job(job_id, job_data)
+        return self._peek_common('ready')
 
     def peek_delayed(self):
         """Peek at the job job on the delayed queue"""
-        with self._sock_ctx() as socket:
-            self._send_message('peek-delayed', socket)
-            job_id, job_data = self._receive_id_and_data_with_prefix(b'FOUND', socket)
-            return Job(job_id, job_data)
+        return self._peek_common('delayed')
 
     def peek_buried(self):
         """Peek at the top job on the buried queue"""
-        with self._sock_ctx() as socket:
-            self._send_message('peek-buried', socket)
-            job_id, job_data = self._receive_id_and_data_with_prefix(b'FOUND', socket)
-            return Job(job_id, job_data)
+        return self._peek_common('buried')
 
     def _common_iter(self, kallable, error):
         while True:
@@ -323,6 +465,10 @@ class BeanstalkClient(object):
         """Reserve jobs as an iterator. Ends iteration when there are no more jobs immediately available"""
         return self._common_iter(lambda: self.reserve_job(0), 'TIMED_OUT')
 
+    def peek_ready_iter(self):
+        """Peek at ready jobs in sequence"""
+        return self._common_iter(self.peek_ready, 'NOT_FOUND')
+
     def peek_delayed_iter(self):
         """Peek at delayed jobs in sequence"""
         return self._common_iter(self.peek_delayed, 'NOT_FOUND')
@@ -355,6 +501,8 @@ class BeanstalkClient(object):
     def release_job(self, job_id, pri=65536, delay=0):
         """Put a job back on the queue to be processed (indicating that you've aborted it)
 
+        You can only release a job which you have reserved using :func:`reserve_job()` or :func:`reserve_iter()`.
+
         :param job_id: Job ID to return
         :param pri: New priority (if not passed, will use old priority)
         :type pri: int
@@ -367,17 +515,54 @@ class BeanstalkClient(object):
             self._send_message('release {0} {1} {2}\r\n'.format(job_id, pri, delay), socket)
             return self._receive_word(socket, b'RELEASED', b'BURIED')
 
+    def kick_job(self, job_id):
+        """Kick the given job id. The job must either be in the DELAYED or BURIED state and will be immediately moved to
+        the READY state."""
+        if hasattr(job_id, 'job_id'):
+            job_id = job_id.job_id
+        with self._sock_ctx() as socket:
+            self._send_message('kick-job {0}'.format(job_id), socket)
+            self._receive_word(socket, b'KICKED')
+
     def use(self, tube):
         """Start producing jobs into the given tube.
 
         :param tube: Name of the tube to USE
 
-        Subsequent calls to :func:put_job()` insert jobs into this tube.
+        Subsequent calls to :func:`put_job` insert jobs into this tube.
         """
         with self._sock_ctx() as socket:
-            self._send_message('use {0}'.format(tube), socket)
-            self._receive_name(socket)
-            self.current_tube = tube
+            if self.current_tube != tube:
+                self.desired_tube = tube
+                self._send_message('use {0}'.format(tube), socket)
+                self._receive_name(socket)
+                self.current_tube = tube
+
+    @contextmanager
+    def using(self, tube):
+        """Context-manager to insert jobs into a specific tube
+
+        :param tube: Tube to insert to
+
+        Yields out an instance of :class:`BeanstalkInsertingProxy` to insert items into that tube
+
+        .. seealso::
+
+           :func:`use()`
+              Change the default tube
+
+           :func:`put_job()`
+              Put a job into whatever the current tube is
+
+           :func:`put_job_into()`
+              Put a job into a specific tube
+        """
+        try:
+            current_tube = self.current_tube
+            self.use(tube)
+            yield BeanstalkInsertingProxy(self, tube)
+        finally:
+            self.use(current_tube)
 
     def kick_jobs(self, num_jobs):
         """Kick some number of jobs from the buried queue onto the ready queue.
